/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* -*-  c-basic-offset: 3 -*- */

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "message.h"
#include "memory.h"
#include "symbol.h"


/* SYM *********************************************************************/

/**< Initialize name of symbol.
 */
static void symInitName(SYM* sym, const char* name)
{
   assert(sym != NULL);

   if( name == NULL )
      sym->name = NULL;
   else
   {
      duplicateMemoryArray(&sym->name, name, strlen(name)+1);
   }
}

/**< Creates atom symbol.
 */
static SYM* symCreateAtom(const char* name)
{
   SYM* sym;

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->firstvarnum = -1;
   sym->numvar = 0;
   sym->type = SYMTYPE_ATOM;

   return sym;
}


/**< Creates binvar symbol.
 */
SYM* symCreateBinvar(const char* name)
{
   SYM* sym;

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->type = SYMTYPE_BINVAR;
   sym->firstvarnum = -1;
   sym->numvar = 1;

   return sym;
}

/**< Creates intvar symbol.
 */
SYM* symCreateIntvar(const char* name, Longint lower, Longint upper)
{
   SYM* sym;

   assert(upper >= lower);

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->type = SYMTYPE_INTVAR;
   sym->firstvarnum = -1;
   sym->numvar = 1;
   sym->value.intvar.lower = lower;
   sym->value.intvar.upper = upper;
   sym->value.intvar.bitarray = NULL;

   return sym;
}

/**< Creates enumvar symbol.
 */
SYM* symCreateEnumvar(const char* name)
{
   SYM* sym;

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->type = SYMTYPE_ENUMVAR;
   sym->firstvarnum = -1;
   sym->numvar = 0;
   sym->value.enumvar.numelem = 0;
   sym->value.enumvar.elemlist = NULL;

   return sym;
}

/**< Adds atom symbol to enum variable's set
 */
int symAddEnumEntry(SYM* enumsym, SYM* atomsym)
{
   int retval;

   assert(enumsym != NULL);
   assert(atomsym != NULL);
   assert(enumsym->type == SYMTYPE_ENUMVAR);
   assert(atomsym->type == SYMTYPE_ATOM);

   retval = symAdd(&(enumsym->value.enumvar.elemlist), atomsym);
   if( retval == SUCCESS )
   {
      enumsym->value.enumvar.numelem++;
      enumsym->numvar++;
   }

   return retval;
}

/**< Creates array symbol.
 */
SYM* symCreateArray(const char* name, int first, int last, SYM* elemtype)
{
   SYM* sym;

   assert(first >= 0);
   assert(last >= first);
   assert(elemtype != NULL);
   assert(elemtype->name == NULL);

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->type = SYMTYPE_ARRAY;
   sym->firstvarnum = -1;
   sym->numvar = (last-first+1) * symGetNumVariables(elemtype);
   sym->value.array.first = first;
   sym->value.array.last = last;
   sym->value.array.elemtype = elemtype;

   return sym;
}

/**< Creates a typeless symbol.
 */
static SYM* symCreateDefine(const char* name, const Node* subnode)
{
   SYM* sym;

   assert(name != NULL);
   assert(subnode != NULL);

   allocMemory(&sym);
   assert(sym != NULL);

   symInitName(sym, name);
   sym->type = SYMTYPE_DEFINE;
   sym->firstvarnum = -1;
   sym->numvar = 0;
   sym->value.define.subnode = subnode;

   return sym;
}


/**< Destroy a symbol and release all associated memory.
 *   Destroying an enumeration symbol does not cause any enumeration atoms
 *   to be destroyed, as these are assumed to be kept in the global symbol
 *   table, nor are any parse trees of DEFINE symbols destroyed.
 */
void symDestroy(SYM* sym)
{
   SYMTABLE* list = NULL;
   SYMTABLE* current = NULL;

   if (sym != NULL) {
      freeMemoryArrayNull(&sym->name);
      switch(sym->type)
      {
      case SYMTYPE_ATOM:
      case SYMTYPE_BINVAR:
      case SYMTYPE_INTVAR:
      case SYMTYPE_DEFINE:
	 break;
      case SYMTYPE_ENUMVAR:
	 list = sym->value.enumvar.elemlist;
	 while (list != NULL) {
	    current = list;
	    list = current->next;
	    freeMemory(&current);
	 }
	 break;
      case SYMTYPE_ARRAY:
	 symDestroy(sym->value.array.elemtype);
	 break;
      default:
	 assert(FALSE);  /* Unknown symbol type! */
	 break;
      }
      freeMemory(&sym);
   }
}


/**< Creates a deep copy of a symbol object.
 *   This method recursively creates copies of the given symbol and all child
 *   nodes of array symbols; enumeration elements are not copied.
 *   @returns Copy of input symbol.
 */
SYM* symCreateDeepCopy(const SYM* source)
{
   SYM* dest;
   SYMTABLE *table;

   assert(source != NULL);

   allocMemory(&dest);
   assert(dest != NULL);

   symInitName(dest, source->name);
   dest->firstvarnum = source->firstvarnum;
   dest->numvar = source->numvar;
   switch( (dest->type = source->type) ) {
   case SYMTYPE_ATOM:
      break;
   case SYMTYPE_BINVAR:
      break;
   case SYMTYPE_INTVAR:
      dest->value.intvar.lower = source->value.intvar.lower;
      dest->value.intvar.upper = source->value.intvar.upper;
      break;
   case SYMTYPE_ENUMVAR:
      dest->value.enumvar.numelem = 0;
      dest->value.enumvar.elemlist = NULL;
      table = source->value.enumvar.elemlist;
      while(table != NULL)
      {
         symAddEnumEntry(dest, table->sym);
         table = table->next;
      }
      assert(dest->value.enumvar.numelem == source->value.enumvar.numelem);
      break;
   case SYMTYPE_ARRAY:
      dest->value.array.first = source->value.array.first;
      dest->value.array.last = source->value.array.last;
      dest->value.array.elemtype = symCreateDeepCopy(source->value.array.elemtype);
      break;
   case SYMTYPE_DEFINE:
      dest->value.define.subnode = source->value.define.subnode;
      break;
   default:
      assert(FALSE);  /* Unknown symbol type! */
      break;
   }

   return dest;
}


/**< Gets name of the symbol.
 */
const char* symGetName(const SYM* sym)
{
   assert(sym != NULL);
   
   return sym->name;
}

/**< Check whether this symbol represents an integer or binary variable.
 */
Bool symIsIntOrBin(const SYM* sym)
{
   assert(sym != NULL);

   switch (sym->type)
   {
   case SYMTYPE_BINVAR:
   case SYMTYPE_INTVAR:
      return TRUE;
   case SYMTYPE_ARRAY:
      return symIsIntOrBin(sym->value.array.elemtype);
   default:
      return FALSE;
   }
}


/**< Check whether this symbol represents an enumeration variable or constant.
 */
Bool symIsEnum(const SYM* sym)
{
   assert(sym != NULL);

   switch (sym->type)
   {
   case SYMTYPE_ATOM:
   case SYMTYPE_ENUMVAR:
      return TRUE;
   case SYMTYPE_ARRAY:
      return symIsEnum(sym->value.array.elemtype);
   default:
      return FALSE;
   }
}


/**< Check, if the bit decomposition of integer variable already exists.
 */
Bool symIsIntvarBitexpanded(const SYM* sym)
{
   assert(sym != NULL);
   assert(sym->type == SYMTYPE_INTVAR);
   
   return( sym->value.intvar.bitarray != NULL );
}

/**< Get the bit decomposition of integer variable.
 */
const SYM* symGetIntvarBitarray(const SYM* sym)
{
   assert(sym != NULL);
   assert(sym->type == SYMTYPE_INTVAR);
   
   return sym->value.intvar.bitarray;
}

/**< Set the bit decomposition of integer variable.
 */
void symSetIntvarBitarray(const SYM* sym, const SYM* bitarray)
{
   assert(sym != NULL);
   assert(sym->type == SYMTYPE_INTVAR);
   
   ((SYM*)sym)->value.intvar.bitarray = bitarray;
}

/**< Get the minimum value of this symbol (or of its array members).
 *   The symbol must be of type binary or integer, or array of
 *   binary or integer.
 */
Longint symGetMinVal(const SYM* sym)
{
   assert(sym != NULL);
   
   switch (sym->type)
   {
   case SYMTYPE_BINVAR:
      return 0;
   case SYMTYPE_INTVAR:
      return sym->value.intvar.lower;
   case SYMTYPE_ARRAY:
      return symGetMinVal(sym->value.array.elemtype);
   default:
      assert(FALSE);  /* Bad symbol type to compute minimum! */
      return 0;
   }
}


/**< Get the maximum value of this symbol (or of its array members).
 *   The symbol must be of type binary or integer, or array of
 *   binary or integer.
 */
Longint symGetMaxVal(const SYM* sym)
{
   assert(sym != NULL);
   
   switch (sym->type)
   {
   case SYMTYPE_BINVAR:
      return 1;
   case SYMTYPE_INTVAR:
      return sym->value.intvar.upper;
   case SYMTYPE_ARRAY:
      return symGetMaxVal(sym->value.array.elemtype);
   default:
      assert(FALSE);  /* Bad symbol type to compute maximum! */
      return 0;
   }
}


/**< Check whether the enumeration type of a symbol contains a given atom.
 *   The given symbol must be an enumeration variable, and the atom must
 *   be an enumeration atom.
 *   @returns true if enumeration contains an atom with the same name
 *            as the specified atom.
 */
Bool symContainsEnumAtom(const SYM* enumsym, const SYM* atom)
{
   assert(enumsym != NULL);
   assert(enumsym->type == SYMTYPE_ENUMVAR);
   assert(atom != NULL);
   assert(atom->type == SYMTYPE_ATOM);

   return symFind(enumsym->value.enumvar.elemlist, atom->name) != NULL;
}


/**< Check whether the enumeration type of a symbol contains another.
 *   Both given symbols must be enumeration variables.
 *   @returns true if enumsym1 contains enumsym2.
 */
Bool symContainsEnumEnum(const SYM* enumsym1, const SYM* enumsym2)
{
   SYMTABLE* list = NULL;

   assert(enumsym1 != NULL);
   assert(enumsym1->type == SYMTYPE_ENUMVAR);
   assert(enumsym2 != NULL);
   assert(enumsym2->type == SYMTYPE_ENUMVAR);

   for (list = enumsym2->value.enumvar.elemlist;
	list != NULL;
	list = list->next) {
      if (!symContainsEnumAtom(enumsym1, list->sym)) {
	 return FALSE;
      }
   }

   return TRUE;
}


/**< Intersect an enumeration symbol type with another.
 *   This method removes all enumeration atoms contained in the element
 *   list of this symbol that are not contained in the element list of
 *   the given other symbol.
 *   Both symbol must be enumeration variables.
 */
void symIntersectEnum(SYM* enumsym, const SYM* other)
{
   SYMTABLE** listref = NULL;
   SYMTABLE* current = NULL;

   assert(enumsym != NULL);
   assert(enumsym->type == SYMTYPE_ENUMVAR);
   assert(other != NULL);
   assert(other->type == SYMTYPE_ENUMVAR);

   listref = &enumsym->value.enumvar.elemlist;
   while (*listref != NULL) {
      current = *listref;
      if (symContainsEnumAtom(other, current->sym)) {
	 /* Element is contained in other symbol: continue ... */
	 listref = &current->next;
      } else {
	 /* Element not contained in other symbol: delete list node ... */
	 *listref = current->next;
	 freeMemory(&current);
      }
   }
}


/**< Unite an enumeration symbol type with another.
 *   This method adds all enumeration atoms contained in the element
 *   list of the given other symbol to the element list of this symbol,
 *   provided that they are not yet contained.
 *   Both symbol must be enumeration variables.
 */
void symUniteEnum(SYM* enumsym, const SYM* other)
{
   SYMTABLE* list = NULL;

   assert(enumsym != NULL);
   assert(enumsym->type == SYMTYPE_ENUMVAR);
   assert(other != NULL);
   assert(other->type == SYMTYPE_ENUMVAR);

   for (list = other->value.enumvar.elemlist;
	list != NULL;
	list = list->next) {
      symAddEnumEntry(enumsym, list->sym);
   }
}


/**< Stores all bound constraints needed for symbol in LP file 'f'.
 */
void symBoundsToLP(const SYM* sym, int maxT, int numvar, FILE* f)
{
   SYM* elemtype;
   int  varnum;
   Longint  lower;
   Longint  upper;
   int  i;
   int  t;

   assert(sym != NULL);
   assert(f != NULL);

   varnum = sym->firstvarnum;

   switch(sym->type)
   {
   case SYMTYPE_ATOM:
      break;
   case SYMTYPE_BINVAR:
      assert(varnum >= 0);
      fprintf(f, "\\ BINVAR: %s\n", sym->name);
      for( t = 0; t <= maxT; ++t )
         fprintf(f, "0 <= x%d <= 1\n", t*numvar + varnum);
      break;
   case SYMTYPE_INTVAR:
      assert(varnum >= 0);
      fprintf(f, "\\ INTVAR: %s\n", sym->name);
      for( t = 0; t <= maxT; ++t )
         fprintf(f, "%lld <= x%d <= %lld\n", sym->value.intvar.lower, t*numvar + varnum, sym->value.intvar.upper);
      break;
   case SYMTYPE_DEFINE:
      break;
   case SYMTYPE_ENUMVAR:
      assert(varnum >= 0);
      assert(sym->numvar == sym->value.enumvar.numelem);
      fprintf(f, "\\ ENUMVAR: %s\n", sym->name);
      for( t = 0; t <= maxT; ++t )
         for( i = 0; i < sym->numvar; ++i )
            fprintf(f, "0 <= x%d <= 1\n", t*numvar + varnum+i);
      break;
   case SYMTYPE_ARRAY:
      elemtype = sym->value.array.elemtype;
      while( elemtype != NULL && elemtype->type == SYMTYPE_ARRAY )
         elemtype = elemtype->value.array.elemtype;
      assert(elemtype != NULL);
      assert(elemtype->type == SYMTYPE_BINVAR
         || elemtype->type == SYMTYPE_INTVAR
         || elemtype->type == SYMTYPE_ENUMVAR);
      if( elemtype->type == SYMTYPE_INTVAR )
      {
         lower = elemtype->value.intvar.lower;
         upper = elemtype->value.intvar.upper;
      }
      else
      {
         lower = 0;
         upper = 1;
      }
      fprintf(f, "\\ ARRAY: %s\n", sym->name);
      for( t = 0; t <= maxT; ++t )
         for( i = 0; i < sym->numvar; ++i )
            fprintf(f, "%lld <= x%d <= %lld\n", lower, t*numvar + varnum+i, upper);
      break;
   default:
      assert(FALSE);  /* Unknown symbol type! */
      break;
   }
}


/* SYMTABLE ****************************************************************/

/**< Finds and returns symbol with name 'name' in symbol table.
 *   @returns NULL, if symbol doesn't exist
 */
SYM* symFind(const SYMTABLE* table, const char* name)
{
   SYM* sym = NULL;

   assert(name != NULL);

   while( table != NULL )
   {
      sym = table->sym;
      if( strcmp(sym->name, name) == 0 )
         return sym;
      table = table->next;
   }
   return NULL;
}


/**< Finds symbol with name 'name' in symbol table.
 *   @returns Position of symbol in table.
 */
int symFindPosition(const SYMTABLE* table, const char* name)
{
   SYM* sym = NULL;
   int index = 0;

   assert(name != NULL);

   while( table != NULL )
   {
      sym = table->sym;
      if( strcmp(sym->name, name) == 0 )
         return index;
      table = table->next;
      index++;
   }
   return -1;
}


/**< Adds symbol to symbol table.
 *   @returns SUCCESS, if successful, FAILURE if symbol name already exists.
 */
int symAdd(SYMTABLE** table, SYM* sym)
{
   SYMTABLE *newtable;
   SYM* findsym;

   assert(table != NULL);
   assert(sym != NULL);

   findsym = symFind(*table, sym->name);
   if( findsym != NULL )
      return FAILURE;

   allocMemory(&newtable);
   assert(newtable != NULL);

   newtable->sym = sym;
   newtable->next = *table;
   *table = newtable;

   return SUCCESS;
}


static Longint symGetConstant(const Node* node)
{
   assert(node != NULL);
   assert(node->type == NODE_NUMBER || node->type == NODE_UMINUS);
   
   if( node->type == NODE_UMINUS )
   {
      assert(node->right == NULL);
      return -symGetConstant(node->left);
   }
   else
   {
      assert(node->right == NULL);
      assert(node->left == NULL);
      return node->value.numb;
   }
}

static SYM* symCreateFromTree(SYMTABLE **table, const char* symname, const Node* typenode)
{
   SYM*  sym = NULL;
   SYM*  atomsym;
   SYM*  elemtype;
   Node* node;
   char  s[255];
   int   restat;
   Longint   leftnumb;
   Longint   rightnumb;

   assert(table != NULL);
   assert(typenode != NULL);

   switch(typenode->type)
   {
   case NODE_BOOLEAN:  /* boolean */
      assert(typenode->left == NULL);
      assert(typenode->right == NULL);
      sym = symCreateBinvar(symname);
      break;
   case NODE_TWODOTS:  /* a..b */
      leftnumb = symGetConstant(typenode->left);
      rightnumb = symGetConstant(typenode->right);
      sym = symCreateIntvar(symname, leftnumb, rightnumb);
      break;
   case NODE_SCALAR:   /* {c1,c2,...} */
      assert(typenode->left != NULL && typenode->left->type == NODE_CONS);
      assert(typenode->right == NULL);

      /* create enum symbol */
      sym = symCreateEnumvar(symname);

      /* create missing enum values, and create list of enum values for enum symbol */
      node = typenode->left;
      while(node != NULL)
      {
         assert(node->type == NODE_CONS);
         assert(node->left != NULL && node->left->type == NODE_ATOM);
         atomsym = symFind(*table, node->left->value.strg);
         if( atomsym != NULL )
         {
            /* enum value already defined in symbol table */
            if( atomsym->type != SYMTYPE_ATOM )
            {
               sprintf(s, "Redefinition of symbol <%s> in line <%d:%d>",
                  node->left->value.strg, node->left->lineno, node->left->column);
               errorMessage(s);
               abort();
            }
         }
         else
         {
            /* create new enum value in symbol table */
            atomsym = symCreateAtom(node->left->value.strg);
            restat = symAdd(table, atomsym);
            assert(restat == SUCCESS);
         }
         assert(atomsym != NULL);
         
         /* add enum value to value list of enum symbol */
         restat = symAddEnumEntry(sym, atomsym);
         if( restat != SUCCESS )
         {
            sprintf(s, "Double enumeration value <%s> in line <%d:%d>",
               atomsym->name, node->left->lineno, node->left->column);
            errorMessage(s);
            abort();
         }

         node = node->right;
      }
      break;
   case NODE_ARRAY:
      assert(typenode->left != NULL && typenode->left->type == NODE_TWODOTS);
      assert(typenode->right != NULL);

      /* retrieve element type of array from right subtree */
      elemtype = symCreateFromTree(table, NULL, typenode->right);
      assert(elemtype != NULL);

      /* retrieve array bounds from left subtree */
      assert(typenode->left->left != NULL && typenode->left->left->type == NODE_NUMBER);
      assert(typenode->left->right != NULL && typenode->left->right->type == NODE_NUMBER);
      sym = symCreateArray(symname, typenode->left->left->value.numb, typenode->left->right->value.numb, elemtype);
      break;
   default:
      sprintf(s, "Invalid nodetype for type node in input line <%d:%d>", typenode->lineno, typenode->column);
      errorMessage(s);
      node_display(typenode);
      abort();
   }

   return sym;
}


/**< Constructs a string representing the symbol name of the node.
 *   The user is resposible to free the memory of the string.
 */
char* symGetNameFromNode(const Node* node)
{
   char* name = NULL;
   char* leftname = NULL;
   char* rightname = NULL;

   assert(node != NULL);
   assert(node->type == NODE_ATOM || node->type == NODE_DOT);
   
   if( node->type == NODE_DOT )
   {
      leftname = symGetNameFromNode(node->left);
      rightname = symGetNameFromNode(node->right);
      allocMemoryArray(&name, strlen(leftname) + strlen(rightname) + 2);
      sprintf(name, "%s.%s", leftname, rightname);
      freeMemoryArray(&leftname);
      freeMemoryArray(&rightname);
   }
   else
   {
      duplicateMemoryArray(&name, node->value.strg, strlen(node->value.strg) + 1);
   }

   return name;
}

/**< Extends symbol table with symbol of given COLON syntax subtree.
 */
static void symExtendTableFromColon(SYMTABLE **table, const Node* node)
{
   SYM*   sym;
   char*  symname;
   int    restat;

   assert(table != NULL);
   assert(node != NULL);
   assert(node->type == NODE_COLON);
   assert(node->left != NULL);

   symname = symGetNameFromNode(node->left);
   assert(symname != NULL);
   
   sym = symCreateFromTree(table, symname, node->right);
   restat = symAdd(table, sym);
   if( restat != SUCCESS )
   {
      char s[255];
      sprintf(s, "Error inserting symbol <%s>", symname);
      errorMessage(s);
      abort();
   }

   freeMemoryArray(&symname);
}

/**< Extends symbol table with all symbols of given VAR or IVAR syntax subtree.
 */
static void symExtendTableFromVar(SYMTABLE **table, const Node* node)
{
   Node* consnode;

   assert(table != NULL);
   assert(node != NULL);
   assert(node->type == NODE_VAR || node->type == NODE_IVAR);

   consnode = node->left; /* left child of VAR/IVAR should be a CONS */
   while(consnode != NULL)
   {
      assert(consnode->type == NODE_CONS);
      symExtendTableFromColon(table, consnode->left);
      consnode = consnode->right;
   }
}

/**< Extends symbol table with symbol of given EQDEF syntax subtree.
 */
static void symExtendTableFromEqdef(SYMTABLE **table, const Node* node)
{
   SYM*         sym;
   const char*  symname;
   int          restat;

   assert(table != NULL);
   assert(node != NULL);
   assert(node->type == NODE_EQDEF);
   assert(node->left != NULL && node->left->type == NODE_ATOM);

   symname = node->left->value.strg;
   assert(symname != NULL);
   
   sym = symCreateDefine(symname, node->right);
   restat = symAdd(table, sym);
   if( restat != SUCCESS )
   {
      char s[255];
      sprintf(s, "Error inserting symbol <%s>", symname);
      errorMessage(s);
      abort();
   }
}

/**< Extends symbol table with all symbols of given DEFINE syntax subtree.
 */
static void symExtendTableFromDefine(SYMTABLE **table, const Node* node)
{
   Node* consnode;

   assert(table != NULL);
   assert(node != NULL);
   assert(node->type == NODE_DEFINE);

   consnode = node->left; /* left child of DEFINE should be a CONS */
   while(consnode != NULL)
   {
      assert(consnode->type == NODE_CONS);
      symExtendTableFromEqdef(table, consnode->left);
      consnode = consnode->right;
   }
}

/**< Extends symbol table with all symbols of given syntax tree.
 */
void symExtendTable(SYMTABLE **table, const Node* node)
{
   assert(table != NULL);
   
   if( node == NULL )
      return;

   switch(node->type)
   {
   case NODE_ERROR: 
      errorMessage("error in node");
      abort();
   case NODE_FALSEEXP: 
   case NODE_TRUEEXP: 
   case NODE_NUMBER: 
   case NODE_ATOM: 
   case NODE_DOT: 
   case NODE_GOTO: 
   case NODE_LET: 
   case NODE_EVAL: 
   case NODE_INIT: 
   case NODE_FAILNESS: 
   case NODE_TRANS: 
   case NODE_CONSTRAINT: 
   case NODE_SIMPWFF: 
   case NODE_CONTEXT: 
   case NODE_CTLWFF: 
   case NODE_LTWFF: 
   case NODE_EQDEF: 
      break;
   case NODE_CONS: 
      symExtendTable(table, node->left);
      symExtendTable(table, node->right);
      break;
   case NODE_VAR: 
   case NODE_IVAR: /* VAR statement */
      symExtendTableFromVar(table, node);
      break;
   case NODE_COLON: 
   case NODE_BOOLEAN: 
   case NODE_SCALAR: 
   case NODE_ARRAY: 
   case NODE_PROCESS: 
   case NODE_TWODOTS: 
   case NODE_UMINUS: 
   case NODE_MODTYPE: 
   case NODE_ASSIGN: 
   case NODE_AND: 
   case NODE_SMALLINIT: 
   case NODE_NEXT: 
   case NODE_INVAR: 
      break;
   case NODE_DEFINE: /* DEFINE statement */
      symExtendTableFromDefine(table, node);
      break;
   case NODE_JUSTICE: 
   case NODE_COMPASSION: 
   case NODE_SPEC: 
   case NODE_LTLSPEC: 
   case NODE_INVARSPEC: 
   case NODE_COMPUTE: 
   case NODE_MINU: 
   case NODE_MAXU: 
   case NODE_SELF: 
   case NODE_IMPLIES: 
   case NODE_IFF: 
   case NODE_OR: 
   case NODE_XOR: 
   case NODE_XNOR: 
   case NODE_NOT: 
   case NODE_PLUS: 
   case NODE_MINUS: 
   case NODE_TIMES: 
   case NODE_IMPLEMENTS: 
   case NODE_INPUT: 
   case NODE_OUTPUT: 
   case NODE_FAIRNESS: 
   case NODE_DIVIDE: 
   case NODE_MOD: 
   case NODE_EQUAL: 
   case NODE_NOTEQUAL: 
   case NODE_LT: 
   case NODE_GT: 
   case NODE_LE: 
   case NODE_GE: 
   case NODE_UNION: 
   case NODE_SETIN: 
   case NODE_CASE: 
   case NODE_EX: 
   case NODE_AX: 
   case NODE_EF: 
   case NODE_AF: 
   case NODE_EG: 
   case NODE_AG: 
   case NODE_AU: 
   case NODE_EU: 
   case NODE_EBF: 
   case NODE_ABF: 
   case NODE_EBG: 
   case NODE_ABG: 
   case NODE_UNTIL: 
   case NODE_SINCE: 
   case NODE_OP_NEXT: 
   case NODE_OP_PREC: 
   case NODE_OP_NOTPRECNOT: 
   case NODE_OP_GLOBAL: 
   case NODE_OP_HISTORICAL: 
   case NODE_OP_FUTURE: 
   case NODE_OP_ONCE: 
   case NODE_LTLWFF: 
   case NODE_COMPWFF: 
      break;
   case NODE_MODULE: 
      assert(node->left != NULL && node->left->type == NODE_MODTYPE);
      symExtendTable(table, node->right);
      break;
   case NODE_ABU: 
   case NODE_EBU:
   case NODE_ISA:
   case NODE_SIGMA:
      break;
   default:
      errorMessage("invalid node type");
      abort();
   }

}

/**< Creates symbol table for given syntax tree.
 */
SYMTABLE* symCreateTable(const Node* root)
{
   SYMTABLE* table = NULL;

   assert(root != NULL);

   symExtendTable(&table, root);

   return table;
}

/**< Displays symbol.
 */
static void symDisplay(const SYM* sym)
{
   SYMTABLE* table;

   assert(sym != NULL);

   if( sym->name != NULL )
      printf("<%s>: ", sym->name);

   if( sym->firstvarnum >= 0 )
   {
      printf("{x%d", sym->firstvarnum);
      if( sym->numvar > 1 )
         printf("..x%d", sym->firstvarnum + sym->numvar - 1);
      printf("} ");
   }

   switch(sym->type)
   {
   case SYMTYPE_ATOM:
      printf("ATOM\n");
      break;
   case SYMTYPE_BINVAR:
      printf("BINVAR\n");
      break;
   case SYMTYPE_INTVAR:
      printf("INTVAR [%lld,%lld]\n", sym->value.intvar.lower, sym->value.intvar.upper);
      break;
   case SYMTYPE_ENUMVAR:
      printf("ENUMVAR [");
      table = sym->value.enumvar.elemlist;
      while(table != NULL)
      {
         assert(table->sym != NULL);
         assert(table->sym->name != NULL);
         assert(table->sym->type == SYMTYPE_ATOM);
         printf(" %s", table->sym->name);
         table = table->next;
      }
      printf(" ]\n");
      break;
   case SYMTYPE_ARRAY:
      printf("ARRAY [%d,%d] of ", sym->value.array.first, sym->value.array.last);
      symDisplay(sym->value.array.elemtype);
      break;
   case SYMTYPE_DEFINE:
      printf("DEFINE\n");
      break;
   default:
      printf("UNKNOWN\n");
      break;
   }
}

/**< Displays symbol table.
 */
void symDisplayTable(const SYMTABLE* table)
{
   while(table != NULL)
   {
      symDisplay(table->sym);
      table = table->next;
   }
}


/**< Gets first MIP variable number of symbol.
 */
int symGetVariableNumber(const SYM* sym)
{
   return sym->firstvarnum;
}

/**< Assigns first MIP variable number 'firstvarnum' to symbol 'sym'.
 */
void symSetVariableNumber(SYM* sym, int firstvarnum)
{
   assert(sym != NULL);

   switch(sym->type)
   {
   case SYMTYPE_ATOM:
   case SYMTYPE_DEFINE:
      sym->firstvarnum = -1;
      break;
   case SYMTYPE_BINVAR:
   case SYMTYPE_INTVAR:
   case SYMTYPE_ENUMVAR:
   case SYMTYPE_ARRAY:
      sym->firstvarnum = firstvarnum;
      break;
   default:
      errorMessage("unknown symbol type");
      abort();
   }
}

/**< Gets the number of variables used for symbol.
 */
int symGetNumVariables(const SYM* sym)
{
   assert(sym != NULL);

   assert( (sym->type == SYMTYPE_ATOM  && sym->numvar == 0)
      || (sym->type == SYMTYPE_BINVAR  && sym->numvar == 1)
      || (sym->type == SYMTYPE_INTVAR  && sym->numvar == 1)
      || (sym->type == SYMTYPE_ENUMVAR && sym->numvar == sym->value.enumvar.numelem)
      || (sym->type == SYMTYPE_ARRAY   && sym->numvar == (sym->value.array.last - sym->value.array.first + 1) *
         symGetNumVariables(sym->value.array.elemtype))
      || (sym->type == SYMTYPE_DEFINE  && sym->numvar == 0) );

   return sym->numvar;
}
