/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef __SYMBOL_H__
#define __SYMBOL_H__

#include "node.h"

enum SymType
{
   SYMTYPE_ATOM     = 0,
   SYMTYPE_BINVAR   = 1,
   SYMTYPE_INTVAR   = 2,
   SYMTYPE_ENUMVAR  = 3,
   SYMTYPE_ARRAY    = 4,
   SYMTYPE_DEFINE   = 5
};
typedef enum SymType SYMTYPE;

typedef struct SymTable SYMTABLE;
typedef struct Sym      SYM;
typedef union  SymValue SYMVALUE;
  
typedef struct SymIntvar SYMINTVAR;
struct SymIntvar
{
   Longint     lower;           /**< lower bound of variable */
   Longint     upper;           /**< upper bound of variable */
   const SYM*  bitarray;        /**< pointer to array variable (if existing) storing the bit decomposition */
};

typedef struct SymEnumvar SYMENUMVAR;
struct SymEnumvar
{
   int         numelem;         /**< number of elements (= number of variables in MIP) */
   SYMTABLE*   elemlist;        /**< available elements for variable */
};

typedef struct SymArray SYMARRAY;
struct SymArray
{
   int         first;           /**< start index of array */
   int         last;            /**< last index of array */
   SYM*        elemtype;        /**< type of array entries, elem->name = NULL */
};

typedef struct SymDefine SYMDEFINE;
struct SymDefine
{
   const Node* subnode;         /**< pointer to subnode representing the define */
};

union SymValue
{
   SYMINTVAR  intvar;
   SYMENUMVAR enumvar;
   SYMARRAY   array;
   SYMDEFINE  define;
};

struct Sym
{
   char*       name;            /**< name of symbol */
   int         firstvarnum;     /**< number of first variable in MIP belonging to symbol */
   int         numvar;          /**< number variables in MIP belonging to symbol */
   SYMTYPE     type;            /**< type of symbol */
   SYMVALUE    value;           /**< value of symbol */
};


struct SymTable
{
   SYM*        sym;
   SYMTABLE*   next;
};


/* SYM */
extern SYM* symCreateBinvar(const char* name);
extern SYM* symCreateIntvar(const char* name, Longint lower, Longint upper);
extern SYM* symCreateEnumvar(const char* name);
extern int symAddEnumEntry(SYM* enumsym, SYM* atomsym);
extern SYM* symCreateArray(const char* name, int first, int last, SYM* elemtype);

extern SYM* symCreateDeepCopy(const SYM* source);
extern void symDestroy(SYM* sym);

extern const char* symGetName(const SYM* sym);
extern Bool symIsIntOrBin(const SYM* sym);
extern Bool symIsEnum(const SYM* sym);
extern Bool symIsIntvarBitexpanded(const SYM* sym);
extern const SYM* symGetIntvarBitarray(const SYM* sym);
extern void symSetIntvarBitarray(const SYM* sym, const SYM* bitarray);
extern Longint symGetMinVal(const SYM* sym);
extern Longint symGetMaxVal(const SYM* sym);
extern Bool symContainsEnumAtom(const SYM* enumsym, const SYM* atom);
extern Bool symContainsEnumEnum(const SYM* enumsym1, const SYM* enumsym2);
extern void symIntersectEnum(SYM* enumsym, const SYM* other);
extern void symUniteEnum(SYM* enumsym, const SYM* other);
extern void symBoundsToLP(const SYM* sym, int maxT, int numvar, FILE* f);

/* SYMTABLE */
extern SYM* symFind(const SYMTABLE* table, const char* name);
extern int symFindPosition(const SYMTABLE* table, const char* name);
extern int symAdd(SYMTABLE** table, SYM* sym);

extern char* symGetNameFromNode(const Node* node);
extern SYMTABLE* symCreateTable(const Node* root);
extern void symDisplayTable(const SYMTABLE* table);

extern int symGetVariableNumber(const SYM* sym);
extern void symSetVariableNumber(SYM* sym, int firstvarnum);
extern int symGetNumVariables(const SYM* sym);

#endif
